import { __awaiter, __decorate } from "tslib";
import { Connect, dataServerResultFilter, DatasourceDefaults, extractFieldDefinitions, FieldTypeEnum, MessageType, toFieldMetadata, } from '@genesislcap/foundation-comms';
import { LifecycleMixin } from '@genesislcap/foundation-utils';
import { attr, customElement, DOM } from '@microsoft/fast-element';
import { FoundationElement } from '@microsoft/fast-foundation';
import { take } from 'rxjs/operators';
import { dateTimeValueFormatter, dateValueFormatter, getColumnType, getFilterByFieldType, getFilterParamsByFieldType, } from '../grid-pro.definitions';
import { GridProRendererTypes } from '../grid-pro.types';
import { logger } from '../utils';
export class StreamDatasource {
    constructor(options) {
        this.rowCount = 0;
        this.moreRows = false;
        this.calculatedRowsCount = 0;
        this.lastSequenceId = 0;
        this.dataserverStream = options.stream;
        this.resourceName = options.resourceName;
        this.resourceParams = options.resourceParams;
        this.originalCriteriaMatch = options.resourceParams.CRITERIA_MATCH;
        this.resourceIndexes = options.resourceIndexes;
        this.resourceColDefs = options.resourceColDefs;
        this.maxRows = options.maxRows;
        this.maxView = options.maxView;
        this.pagination = options.pagination;
    }
    getRows(params) {
        var _a;
        return __awaiter(this, void 0, void 0, function* () {
            const toBeAppliedFilterModel = params.request.filterModel;
            if (JSON.stringify(toBeAppliedFilterModel) !== JSON.stringify(this.currentFilterModel) ||
                Object.keys(toBeAppliedFilterModel).length > 0) {
                if (Object.keys(toBeAppliedFilterModel).length === 0 && this.currentFilterModel) {
                    this.currentFilterModel = null;
                    this.resourceParams.CRITERIA_MATCH = this.originalCriteriaMatch;
                    this.restartSubscription();
                }
                else if (Object.keys(toBeAppliedFilterModel).length > 0) {
                    this.currentFilterModel = toBeAppliedFilterModel;
                    this.resourceParams.CRITERIA_MATCH = this.buildCriteriaMatchFromFilters();
                    this.restartSubscription();
                }
            }
            const toBeAppliedSortModel = params.request.sortModel;
            if (toBeAppliedSortModel.length !== ((_a = this.currentSortModel) === null || _a === void 0 ? void 0 : _a.length) ||
                toBeAppliedSortModel.length > 0) {
                if (toBeAppliedSortModel.length === 0 && this.currentSortModel) {
                    this.currentSortModel = null;
                    this.resourceParams.ORDER_BY = null;
                    this.resourceParams.REVERSE = null;
                    this.restartSubscription();
                }
                else if (toBeAppliedSortModel.length > 0) {
                    this.currentSortModel = toBeAppliedSortModel;
                    const coldIdBeingSorted = this.currentSortModel[0].colId;
                    const sortTypeBeingApplied = this.currentSortModel[0].sort;
                    const orderByAndToBeSortedColIds = this.getOrderByAndToBeSortedColIds(this.resourceIndexes, coldIdBeingSorted);
                    orderByAndToBeSortedColIds.toBeSorted.forEach((colId) => {
                        this.currentSortModel.push({ sort: sortTypeBeingApplied, colId });
                    });
                    this.resourceParams.ORDER_BY = orderByAndToBeSortedColIds.orderBy;
                    this.resourceParams.REVERSE = sortTypeBeingApplied === 'desc' ? true : false;
                    this.restartSubscription();
                }
            }
            if (this.moreRows && params.request.startRow >= this.maxRows) {
                this.connect.getMoreRows(this.streamSourceRef);
            }
            this.dataserverStream.pipe(take(1)).subscribe((result) => {
                var _a, _b;
                const messageType = result.MESSAGE_TYPE;
                if (messageType &&
                    (messageType === MessageType.LOGOFF_ACK || messageType === MessageType.MSG_NACK)) {
                    params.fail();
                    return;
                }
                this.moreRows = result.MORE_ROWS;
                let rowData = [];
                if (result.ROW) {
                    const nextMessage = dataServerResultFilter(result);
                    rowData = this.handleStreamUpdate(nextMessage);
                }
                this.lastSequenceId = result.SEQUENCE_ID;
                if (this.lastSequenceId === 1) {
                    this.streamSourceRef = result.SOURCE_REF;
                    this.rowCount = (_a = result.ROWS_COUNT) !== null && _a !== void 0 ? _a : rowData.length;
                }
                this.calculatedRowsCount += (_b = rowData === null || rowData === void 0 ? void 0 : rowData.length) !== null && _b !== void 0 ? _b : 0;
                const successRowData = { rowData };
                if (this.pagination) {
                    successRowData.rowCount = Math.min(this.rowCount, this.maxView);
                }
                else if (params.request.endRow > this.maxView ||
                    params.request.endRow > this.calculatedRowsCount) {
                    successRowData.rowCount = this.calculatedRowsCount;
                }
                params.success(successRowData);
            });
        });
    }
    restartSubscription() {
        this.connect.dataLogoff(this.streamSourceRef);
        const streamOnMessage = (message) => message;
        const onError = (error) => logger.error(error);
        this.dataserverStream = this.connect.stream(this.resourceName, streamOnMessage, onError, this.resourceParams);
    }
    getOrderByAndToBeSortedColIds(map, searchValue) {
        for (const [key, value] of map.entries()) {
            const searchIndex = value.findIndex((x) => x === searchValue);
            if (searchIndex !== -1) {
                const copy = [...value];
                copy.splice(searchIndex, 1);
                return { orderBy: key, toBeSorted: copy };
            }
        }
    }
    handleStreamUpdate(result) {
        var _a;
        if (!result)
            return;
        const rows = new Map();
        (_a = result.inserts) === null || _a === void 0 ? void 0 : _a.forEach((insertData) => {
            rows.set(insertData.ROW_REF, insertData);
        });
        const rowData = Array.from(rows.values());
        return rowData;
    }
    buildCriteriaMatchFromFilters() {
        this.resourceParams.CRITERIA_MATCH = this.originalCriteriaMatch;
        if (this.resourceParams.CRITERIA_MATCH &&
            typeof this.resourceParams.CRITERIA_MATCH === 'string') {
            this.resourceParams.CRITERIA_MATCH = this.resourceParams.CRITERIA_MATCH.split(';');
        }
        if (!this.resourceParams.CRITERIA_MATCH) {
            this.resourceParams.CRITERIA_MATCH = [];
        }
        const criteria = this.resourceParams.CRITERIA_MATCH.concat(this.criteriaFromFilters());
        return criteria.join(' && ');
    }
    criteriaFromFilters() {
        const filters = [];
        this.getFiltersByType('text').forEach((k) => filters.push(`Expr.containsIgnoreCase(${k}, '${this.currentFilterModel[k].filter.toLowerCase()}')`));
        return filters;
    }
    getFiltersByType(filterType) {
        return Object.keys(this.currentFilterModel)
            .filter((k) => this.resourceColDefs.findIndex((o) => o.NAME === k) !== -1)
            .filter((k) => this.currentFilterModel[k].filterType === filterType);
    }
}
__decorate([
    Connect
], StreamDatasource.prototype, "connect", void 0);
let GridProGenesisDatasourceNext = class GridProGenesisDatasourceNext extends LifecycleMixin(FoundationElement) {
    constructor() {
        super(...arguments);
        this.isSnapshot = false;
        this.maxRows = DatasourceDefaults.MAX_ROWS_250;
        this.maxView = DatasourceDefaults.MAX_VIEW_1000;
        this.pagination = false;
        this.reverse = false;
    }
    connectedCallback() {
        super.connectedCallback();
        DOM.queueUpdate(() => {
            if (!this.shouldRunConnect)
                return;
            this.init();
        });
    }
    disconnectedCallback() {
        super.disconnectedCallback();
        DOM.queueUpdate(() => {
            if (!this.shouldRunDisconnect)
                return;
            this.reset();
        });
    }
    init() {
        if (this.agGrid && this.resourceName && this.connect.isConnected) {
            this.agGrid.gridOptions = {
                cacheBlockSize: this.maxRows,
                defaultColDef: {
                    filter: true,
                    resizable: true,
                    sortable: true,
                },
                pagination: this.pagination,
                paginationPageSize: this.maxRows,
                rowBuffer: 0,
                rowModelType: 'serverSide',
                suppressServerSideInfiniteScroll: false,
                suppressMultiSort: true,
                onFilterChanged: (params) => { },
                onSortChanged: (params) => { },
            };
            this.agGrid.addEventListener('onGridReady', () => __awaiter(this, void 0, void 0, function* () {
                const meta = yield this.connect
                    .getMetadata(this.resourceName)
                    .catch((e) => logger.error(e));
                if (!meta)
                    return false;
                this.indexes = this.getResourceIndexes(meta.INDEXES);
                this.originalFieldDef = extractFieldDefinitions(meta);
                const fieldMetadata = toFieldMetadata(this.originalFieldDef);
                const agColumnDefs = this.getAgColumnDefs(fieldMetadata);
                this.agGrid.gridApi.setColumnDefs(agColumnDefs);
                const dataserverStream = yield this.createDataserverStream();
                this.dataserverStream = new StreamDatasource({
                    stream: dataserverStream,
                    resourceName: this.resourceName,
                    resourceParams: this.params,
                    resourceIndexes: this.indexes,
                    resourceColDefs: this.originalFieldDef,
                    maxRows: this.maxRows,
                    maxView: this.maxView,
                    pagination: this.pagination,
                });
                this.agGrid.gridApi.setServerSideDatasource(this.dataserverStream);
            }), { once: true });
            return;
        }
        logger.warn(`Application not connected, falling back to local columnDefs/rowData`);
    }
    reset() { }
    get agGrid() {
        return this.parentElement;
    }
    getResourceIndexes(avaialbleIndexes) {
        const resourceIndexesMap = new Map();
        avaialbleIndexes === null || avaialbleIndexes === void 0 ? void 0 : avaialbleIndexes.forEach((index) => {
            resourceIndexesMap.set(index.NAME, index.FIELDS.split(' '));
        });
        return resourceIndexesMap;
    }
    get params() {
        const params = {
            CRITERIA_MATCH: this.criteria,
            FIELDS: this.fields,
            MAX_ROWS: this.maxRows,
            MAX_VIEW: this.maxView,
            REVERSE: this.reverse,
        };
        const keys = Array.from(this.indexes.keys());
        if (this.orderBy) {
            if (keys.findIndex((key) => key === this.orderBy) > -1) {
                params.ORDER_BY = this.orderBy;
            }
            else {
                const validIndexes = keys.map((key) => key);
                logger.warn('Specified [orderBy] index is not valid and will not be used. See https://learn.genesis.global/docs/database/data-types/index-entities/ - Available indexes:', validIndexes);
            }
        }
        logger.debug('Getting configured params:', params);
        return params;
    }
    createDataserverStream(existingParams = null) {
        return __awaiter(this, void 0, void 0, function* () {
            const streamOnMessage = (message) => message;
            const onError = (error) => logger.error(error);
            return this.connect.stream(this.resourceName, streamOnMessage, onError, existingParams !== null && existingParams !== void 0 ? existingParams : this.params);
        });
    }
    getAgColumnDefs(fieldsMetadata) {
        const colDefsFromGenesisData = fieldsMetadata === null || fieldsMetadata === void 0 ? void 0 : fieldsMetadata.flatMap((field) => {
            var _a;
            const cellConfigs = {};
            cellConfigs.filterParams = getFilterParamsByFieldType(field.type);
            switch (field.type) {
                case FieldTypeEnum.BOOLEAN:
                    if (this.agGrid.autoCellRendererByType) {
                        cellConfigs.cellRenderer = GridProRendererTypes.boolean;
                    }
                    break;
                case FieldTypeEnum.DATE:
                    cellConfigs.valueFormatter = dateValueFormatter;
                    break;
                case FieldTypeEnum.DATETIME:
                    cellConfigs.valueFormatter = dateTimeValueFormatter;
                    break;
                default:
                    break;
            }
            const templateColDef = (_a = this.agGrid.gridApi) === null || _a === void 0 ? void 0 : _a.getColumnDef(field.name);
            if (this.agGrid.onlyTemplateColDefs && !templateColDef) {
                return [];
            }
            return Object.assign(Object.assign({ field: field.name, type: getColumnType(field.type), filter: getFilterByFieldType(field.type), sortable: [].concat(...this.indexes.values()).includes(field.name) }, cellConfigs), templateColDef);
        });
        const colDefsMergedWithTemplateDefs = this.agGrid.applyTemplateDefinitions(colDefsFromGenesisData);
        return colDefsMergedWithTemplateDefs;
    }
};
__decorate([
    Connect
], GridProGenesisDatasourceNext.prototype, "connect", void 0);
__decorate([
    attr
], GridProGenesisDatasourceNext.prototype, "criteria", void 0);
__decorate([
    attr
], GridProGenesisDatasourceNext.prototype, "fields", void 0);
__decorate([
    attr({ mode: 'boolean', attribute: 'is-snapshot' })
], GridProGenesisDatasourceNext.prototype, "isSnapshot", void 0);
__decorate([
    attr({ attribute: 'max-rows' })
], GridProGenesisDatasourceNext.prototype, "maxRows", void 0);
__decorate([
    attr({ attribute: 'max-view' })
], GridProGenesisDatasourceNext.prototype, "maxView", void 0);
__decorate([
    attr({ attribute: 'order-by' })
], GridProGenesisDatasourceNext.prototype, "orderBy", void 0);
__decorate([
    attr({ attribute: 'resource-name' })
], GridProGenesisDatasourceNext.prototype, "resourceName", void 0);
__decorate([
    attr({ mode: 'boolean' })
], GridProGenesisDatasourceNext.prototype, "pagination", void 0);
__decorate([
    attr({ mode: 'boolean' })
], GridProGenesisDatasourceNext.prototype, "reverse", void 0);
GridProGenesisDatasourceNext = __decorate([
    customElement({
        name: 'grid-pro-genesis-datasource-next',
    })
], GridProGenesisDatasourceNext);
export { GridProGenesisDatasourceNext };
